/*
 *	$Id: enable_bridges.c,v 1.41 2004/05/11 15:39:24 mj Exp $
 *	Modified lspci code to enable bridges behind cardbus
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>

#include "pciutils.h"
#define MAX 255

/* Options */

static int show_hex;			/* Show contents of config space as hexadecimal numbers */
static struct pci_filter filter;	/* Device filter */
static int show_tree;			/* Show bus tree */
static int magma_found = 0;			/* Bus mapping mode enabled */
static int num_cb_found = 0;			/* Bus mapping mode enabled */
static int UNDER_CB = 0;			/* Bus mapping mode enabled */

/* Communication with libpci */

static struct pci_access *pacc;

/*
 *  If we aren't being compiled by GCC, use malloc() instead of alloca().
 *  This increases our memory footprint, but only slightly since we don't
 *  use alloca() much.
 */

#ifndef __GNUC__
#define alloca malloc
#endif

/* Our view of the PCI bus */


struct cb {
	byte cb_bus;
	byte s_bus;
	word io_base1;
	word io_limit1;
	word io_base2;
	word io_limit2;
	int found;
	int mem_base;
	int mem_limit;
	int pref_base;
	int pref_limit;
} cb[MAX];

struct dec {
	byte p_bus;
	byte s_bus;
	byte io_base;
	byte io_limit;
	word mem_base;
	word mem_limit;
	word pref_base;
	word pref_limit;
} dec[MAX];

struct device {
  struct device *next;
  struct pci_dev *dev;
  unsigned int config_cnt;
  byte config[256];
};

static struct device *first_dev;

static struct device *
scan_device(struct pci_dev *p)
{
  int how_much = (show_hex > 2) ? 256 : 64;
  struct device *d;

  if (!pci_filter_match(&filter, p))
    return NULL;
  d = xmalloc(sizeof(struct device));
  bzero(d, sizeof(*d));
  d->dev = p;
  if (!pci_read_block(p, 0, d->config, how_much))
    die("Unable to read %d bytes of configuration space.", how_much);
  if (how_much < 128 && (d->config[PCI_HEADER_TYPE] & 0x7f) == PCI_HEADER_TYPE_CARDBUS)
    {
      /* For cardbus bridges, we need to fetch 64 bytes more to get the full standard header... */
      if (!pci_read_block(p, 64, d->config+64, 64))
	die("Unable to read cardbus bridge extension data.");
      how_much = 128;
    }
  d->config_cnt = how_much;
  pci_setup_cache(p, d->config, d->config_cnt);
  pci_fill_info(p, PCI_FILL_IDENT | PCI_FILL_IRQ | PCI_FILL_BASES | PCI_FILL_ROM_BASE | PCI_FILL_SIZES);
  return d;
}

static void
scan_devices(void)
{
  struct device *d;
  struct pci_dev *p;

  pci_scan_bus(pacc);
  for(p=pacc->devices; p; p=p->next)
    if (d = scan_device(p))
      {
	d->next = first_dev;
	first_dev = d;
      }
}

/* Config space accesses */

static inline byte
get_conf_byte(struct device *d, unsigned int pos)
{
  return d->config[pos];
}

static word
get_conf_word(struct device *d, unsigned int pos)
{
  return d->config[pos] | (d->config[pos+1] << 8);
}

static u32
get_conf_long(struct device *d, unsigned int pos)
{
  return d->config[pos] |
    (d->config[pos+1] << 8) |
    (d->config[pos+2] << 16) |
    (d->config[pos+3] << 24);
}

/* Sorting */

static int
compare_them(const void *A, const void *B)
{
  const struct pci_dev *a = (*(const struct device **)A)->dev;
  const struct pci_dev *b = (*(const struct device **)B)->dev;

  if (a->bus < b->bus)
    return -1;
  if (a->bus > b->bus)
    return 1;
  if (a->dev < b->dev)
    return -1;
  if (a->dev > b->dev)
    return 1;
  if (a->func < b->func)
    return -1;
  if (a->func > b->func)
    return 1;
  return 0;
}

static void
sort_them(void)
{
  struct device **index, **h, **last_dev;
  int cnt;
  struct device *d;

  cnt = 0;
  for(d=first_dev; d; d=d->next)
    cnt++;
  h = index = alloca(sizeof(struct device *) * cnt);
  for(d=first_dev; d; d=d->next)
    *h++ = d;
  qsort(index, cnt, sizeof(struct device *), compare_them);
  last_dev = &first_dev;
  h = index;
  while (cnt--)
    {
      *last_dev = *h;
      last_dev = &(*h)->next;
      h++;
    }
  *last_dev = NULL;
}


/* Tree output */

struct bridge {
  struct bridge *chain;			/* Single-linked list of bridges */
  struct bridge *next, *child;		/* Tree of bridges */
  struct bus *first_bus;		/* List of busses connected to this bridge */
  unsigned int primary, secondary, subordinate;	/* Bus numbers */
  struct device *br_dev;
};

struct bus {
  unsigned int number;
  struct bus *sibling;
  struct device *first_dev, **last_dev;
};

static struct bridge host_bridge = { NULL, NULL, NULL, NULL, ~0, 0, ~0, NULL };

static struct bus *
find_bus(struct bridge *b, unsigned int n)
{
  struct bus *bus;

  for(bus=b->first_bus; bus; bus=bus->sibling)
    if (bus->number == n)
      break;
  return bus;
}

static struct bus *
new_bus(struct bridge *b, unsigned int n)
{
  struct bus *bus = xmalloc(sizeof(struct bus));

  bus = xmalloc(sizeof(struct bus));
  bus->number = n;
  bus->sibling = b->first_bus;
  bus->first_dev = NULL;
  bus->last_dev = &bus->first_dev;
  b->first_bus = bus;
  return bus;
}

static void
insert_dev(struct device *d, struct bridge *b)
{
  struct pci_dev *p = d->dev;
  struct bus *bus;

  if (! (bus = find_bus(b, p->bus)))
    {
      struct bridge *c;
      for(c=b->child; c; c=c->next)
	if (c->secondary <= p->bus && p->bus <= c->subordinate)
          {
            insert_dev(d, c);
            return;
          }
      bus = new_bus(b, p->bus);
    }
  /* Simple insertion at the end _does_ guarantee the correct order as the
   * original device list was sorted by (bus, devfn) lexicographically
   * and all devices on the new list have the same bus number.
   */
  *bus->last_dev = d;
  bus->last_dev = &d->next;
  d->next = NULL;
}

static void
grow_tree(void)
{
  struct device *d, *d2;
  struct bridge **last_br, *b;

  /* Build list of bridges */
  last_br = &host_bridge.chain;
  for(d=first_dev; d; d=d->next)
    {
      word class = get_conf_word(d, PCI_CLASS_DEVICE);
      byte ht = get_conf_byte(d, PCI_HEADER_TYPE) & 0x7f;
//printf("ht  %x \n", ht);
      if (class == PCI_CLASS_BRIDGE_PCI &&
	  (ht == PCI_HEADER_TYPE_BRIDGE || ht == PCI_HEADER_TYPE_CARDBUS))
	{
	  b = xmalloc(sizeof(struct bridge));
	  if (ht == PCI_HEADER_TYPE_BRIDGE)
	    {
	      b->primary = get_conf_byte(d, PCI_CB_PRIMARY_BUS);
	      b->secondary = get_conf_byte(d, PCI_CB_CARD_BUS);
	      b->subordinate = get_conf_byte(d, PCI_CB_SUBORDINATE_BUS);
//printf("BRIDGE  secondary %x subord %x \n", b->secondary, b->subordinate);
	    }
	  else
	    { //cardbus?
	      b->primary = get_conf_byte(d, PCI_PRIMARY_BUS);
	      b->secondary = get_conf_byte(d, PCI_SECONDARY_BUS);
	      b->subordinate = get_conf_byte(d, PCI_SUBORDINATE_BUS);
printf("CARDBUS  secondary %x subord %x \n", b->secondary, b->subordinate);
	    }
	  *last_br = b;
	  last_br = &b->chain;
	  b->next = b->child = NULL;
	  b->first_bus = NULL;
	  b->br_dev = d;
	}
    }
  *last_br = NULL;

  /* Create a bridge tree */

  for(b=&host_bridge; b; b=b->chain)
    {
      struct bridge *c, *best;
      best = NULL;
      for(c=&host_bridge; c; c=c->chain)
	if (c != b && b->primary >= c->secondary && b->primary <= c->subordinate &&
	    (!best || best->subordinate - best->primary > c->subordinate - c->primary))
	  best = c;
      if (best)
	{
	  b->next = best->child;
	  best->child = b;
	}
    }

  /* Insert secondary bus for each bridge */

  for(b=&host_bridge; b; b=b->chain)
    if (!find_bus(b, b->secondary))
      new_bus(b, b->secondary);

  /* Create bus structs and link devices */

  for(d=first_dev; d;)
    {
      d2 = d->next;
      insert_dev(d, &host_bridge);
      d = d2;
    }
}

static void
show_tree_bridge(struct bridge *);

//this needs some work to enable nested bridges below ours
static void
show_tree_dev(struct device *d)
{
  struct pci_dev *q = d->dev;
  struct bridge *b;
  int temp, i;
  //char namebuf[256];

  for(b=&host_bridge; b; b=b->chain){
    if (b->br_dev == d)
      {
	  word vendor_id = get_conf_word(d, 0x00);
	  word device_id = get_conf_word(d, 0x02);
	  if ( (vendor_id == 0x01011 ) || ( vendor_id == 0x3388 ) || (vendor_id == 0x12d8) || (UNDER_CB >= 1 )  ){
		for(i=0;i<=num_cb_found; i++) {
		if ( b->primary == cb[i].cb_bus ){
		  magma_found = i ;
		  magma_found |= 0x0100 ; //FOUND BIT
		  dec[0].p_bus = get_conf_byte (d,0x18);
		  dec[0].s_bus = get_conf_byte (d,0x19);
		 
		  printf ("Enabled bridge %x:%x \n", vendor_id, device_id);
  		  pci_write_word (d->dev , 0x04, 0x0007);
		  pci_write_byte (d->dev , 0x0c, 0x10);
		  pci_write_byte (d->dev , 0x0d, 0xf8);
/*
		  magma_found = magma_found - 0x0100;
		  printf ("magmafound %x\n", magma_found  );
		   temp = cb[magma_found].io_base1 >> 8; 
		   pci_write_byte (d->dev , 0x1c, temp);
		   temp = (cb[magma_found].io_limit2 >> 8) ; 
		   pci_write_byte (d->dev , 0x1d, temp );
		  printf ( "temp %x\n", temp  );
		   temp = cb[magma_found].mem_base >> 16; 
	 	   pci_write_word (d->dev , 0x20, temp);
		   temp = cb[magma_found].mem_limit >> 16; 
		   pci_write_word (d->dev , 0x22, temp);
		   temp = cb[magma_found].pref_base >> 16; 
		   pci_write_word (d->dev , 0x24, temp);
		   temp = cb[magma_found].pref_limit >> 16; 
		   pci_write_word (d->dev , 0x26, temp);

		  magma_found |= 0x0100 ; //FOUND BIT
*/
		  break;
		}
		else if ( (magma_found >= 0x0100)  ) {
		  magma_found = magma_found - 0x0100;
		  printf ("Enabled 2nd bridge %x:%x \n", vendor_id, device_id);
  		   pci_write_word (d->dev , 0x04, 0x0007);
		   pci_write_byte (d->dev , 0x0c, 0x10);
		   pci_write_byte (d->dev , 0x0d, 0xf8);
		   temp = cb[magma_found].io_base1 >> 8; 
		   pci_write_byte (d->dev , 0x1c, temp);
		   temp = (cb[magma_found].io_limit2 >> 8) ; 
		   pci_write_byte (d->dev , 0x1d, temp );
		   temp = cb[magma_found].mem_base >> 16; 
	 	   pci_write_word (d->dev , 0x20, temp);
		   temp = cb[magma_found].mem_limit >> 16; 
		   pci_write_word (d->dev , 0x22, temp);
		   temp = cb[magma_found].pref_base >> 16; 
		   pci_write_word (d->dev , 0x24, temp);
		   temp = cb[magma_found].pref_limit >> 16; 
		   pci_write_word (d->dev , 0x26, temp);
		  dec[1].p_bus = get_conf_byte (d,0x18);
		  dec[1].s_bus = get_conf_byte (d,0x19);
		  UNDER_CB++;
		  break;
		}
		else if ( UNDER_CB >=1  ) {
		  printf ("Enabled 3rd bridge %x:%x \n", vendor_id, device_id);
  		   pci_write_word (d->dev , 0x04, 0x0007);
		   pci_write_byte (d->dev , 0x0c, 0x10);
		   pci_write_byte (d->dev , 0x0d, 0xf8);
		   temp = cb[magma_found].io_base1 >> 8; 
		   pci_write_byte (d->dev , 0x1c, temp);
		   temp = (cb[magma_found].io_limit2 >> 8) ; 
		   pci_write_byte (d->dev , 0x1d, temp );
		   temp = (cb[magma_found].mem_base >> 16) ; 
	 	   pci_write_word (d->dev , 0x20, temp);
		   temp = (cb[magma_found].mem_limit >> 16) - 0x0020; 
		   pci_write_word (d->dev , 0x22, temp);
		   temp = cb[magma_found].pref_base >> 16; 
		   pci_write_word (d->dev , 0x24, temp);
		   temp = (cb[magma_found].pref_limit >> 16) - 0x0020; 
		  // printf (" pre.limit %x \n", temp);
		   pci_write_word (d->dev , 0x26, temp  );
		  dec[1].p_bus = get_conf_byte (d,0x18);
		break;
		}
//	printf("bottom of for\n");
	} //bottom of for
//	printf("out of for\n");
	} //end of if found magma  
        show_tree_bridge(b);
        return;
      }//if b->br
      }//for
if (magma_found > 0 ) {//enable cards in chassis as well as the bridges
    //printf("  %s what's this\nnamebuf %s \n",
	//	 pci_lookup_name(pacc, namebuf, sizeof(namebuf),
	//			 PCI_LOOKUP_VENDOR | PCI_LOOKUP_DEVICE,
	//			 q->vendor_id, q->device_id, 0, 0), namebuf);

   printf ("Enabled device %x:%x %d\n", q->vendor_id, q->device_id, magma_found);
   pci_write_word (q , 0x04, 0x0007);
   pci_write_byte (q , 0x0c, 0x10);
   pci_write_byte (q , 0x0d, 0xf8);
}
			
}

static void
show_tree_bus(struct bus *b)
{ 
  if (!b->first_dev)
	;
  else if (!b->first_dev->next)
    {
//	printf ("1st bus %d \n", magma_found);
      show_tree_dev(b->first_dev);
	magma_found = 0;
    }
  else
    {
      struct device *d = b->first_dev;
      while (d->next)
	{
//	printf ("2nd bus %d \n", magma_found);
	  show_tree_dev(d);
	  d = d->next;
	}
	
//	printf ("3rd bus\n");
      show_tree_dev(d);
      UNDER_CB = 0;
    }
}

static void
show_tree_bridge(struct bridge *b) 
{

  if (!b->first_bus->sibling)
    {
      if (b == &host_bridge)
	printf("HOST BRIDGE TRUE\n");
      show_tree_bus(b->first_bus);
    }
  else
    {
	
      struct bus *u = b->first_bus;

      while (u->sibling)
        {
//	printf("in tree bus else\n");
          show_tree_bus(u);
          u = u->sibling;
        }
      show_tree_bus(u);
    }

}


static void fill_cb_info(struct device *d, int i)
{
	word brc;

	cb[i].found = i;
	cb[i].cb_bus =  get_conf_byte(d, 0x19); //jme
	cb[i].s_bus =  get_conf_byte(d, 0x1a); //jme
	cb[i].io_base1 =  get_conf_word(d, 0x2c); //jme
	cb[i].io_limit1 =  get_conf_word(d, 0x30); //jme
	cb[i].io_base2 =  get_conf_word(d, 0x34); //jme
	cb[i].io_limit2 =  get_conf_word(d, 0x38); //jme

  	brc = get_conf_word(d, PCI_BRIDGE_CONTROL);

	if ( brc && 0x100) {
	  cb[i].mem_base =  get_conf_long(d, 0x24); //jme
	  cb[i].mem_limit =  get_conf_long(d, 0x28); //jme
	  cb[i].pref_base =  get_conf_long(d, 0x1c); //jme
	  cb[i].pref_limit =  get_conf_long(d, 0x20); //jme
	}
	else {
	  cb[i].mem_base =  get_conf_long(d, 0x1c); //jme
	  cb[i].mem_limit =  get_conf_long(d, 0x20); //jme
	  cb[i].pref_base =  get_conf_long(d, 0x24); //jme
	  cb[i].pref_limit =  get_conf_long(d, 0x28); //jme
	}
	
}



static void
show_forest(void)
{
  struct device *d;
  int  i = 1;
  word x;

  grow_tree();
//jme
  cb[0].found = 0;
  for(d=first_dev; d; d=d->next){
 	x = get_conf_byte(d, 0x0a); // Get Sub class code to know if this is cardbus
	if ( x == 0x07){
	   fill_cb_info(d, i);
	   i++;
	}
  }
  num_cb_found = i - 1;
//jme
  show_tree_bridge(&host_bridge);
}


int main(int argc, char **argv)
{
  if (argc == 2 && !strcmp(argv[1], "--version"))
    {
      puts("enable_bridges version " PCIUTILS_VERSION);
      return 0;
    }

  pacc = pci_alloc();
  pacc->error = die;
  pci_filter_init(pacc, &filter);

  show_tree++;

  pci_init(pacc);
  scan_devices();
  sort_them();
  show_forest();
  pci_cleanup(pacc);

  return 0;
}
